using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using DWG;
using RasterImage;
using System.IO;
using System.Drawing.Imaging;
using System.Collections;
using Microsoft.WindowsCE.Forms;
using System.Windows.Forms;
using CADImport;
using HPGL2;

namespace CompactViewer
{
    internal struct ScaleData
    {
        public Point position;
        public float scale;
    }

    public partial class MainForm : Form
    {
        internal static CADImage FCADImage;
        private float scale;
        private float prev_scale = 1;
        private Point pos;
        private Orbit3D Orb3D = new Orbit3D();
        private float wh;
        private Rectangle curClRect;
        private Point old_Pos = new Point();
        private int cX, cY;
        private bool det1;
        private AboutForm abForm1;
        private RegForm regForm1;
        private bool deactiv;
        internal static SortedList cadImageList;
        internal static string curFileName;
        private SortedList scaleCoeff;
        public string argFileName = null;
        private bool onReadLine;
        private ArrayList rlPoints = new ArrayList();
        private ArrayList rlObjects = new ArrayList();
        private DPoint ScaleRect = new DPoint();

        private readonly ArrayList HPGLExt = new ArrayList(new string[]{".HPG", ".PLT", ".RTL", ".SPL",
                                                           ".PRN", ".GL2", ".HPGL2", ".HPGL",
                                                           ".HP2", ".HP1", ".HP", ".PLO",
                                                           ".HG", ".HGL"});

        public MainForm()
        {
            InitializeComponent();
            Initial3DOrbit();
            abForm1 = new AboutForm();
            regForm1 = new RegForm();
            cadImageList = new SortedList();
            curFileName = "";
            scaleCoeff = new SortedList();
        }

        public void Initial3DOrbit()
        {
            Orb3D.Parent = this;
        }

        private void miOpen_Click(object sender, EventArgs e)
        {
            prgProgress.Visible = true;
            string tmp = this.Text;
            this.Text = "Load file...";
            LoadFile(true);
            this.Text = tmp;
            prgProgress.Visible = false;
        }

        public void LoadFile(bool dlg)
        {
            if (dlg)
                if (openFileDlg.ShowDialog() != DialogResult.OK)
                {
                    return;
                }
            this.Invalidate();
            if (openFileDlg.FileName != null)
            {
                if (FCADImage != null)
                {
                    AddNewCADImage();
                }
                scale = 1;
                prev_scale = 1;
                pos = new Point();
                //CADImage.isImageDraw = true;
                Cursor.Current = Cursors.WaitCursor;
                string ext = Path.GetExtension(openFileDlg.FileName).ToUpper();
                if(ext == ".DWG")
                {
                    FCADImage = new DWG.DWGImage();
                    LoadFile(openFileDlg.FileName, true);
                }
                else
                {
                    if (ext == ".DXF")
                    {
                        FCADImage = new CADImage();
                        LoadFile(openFileDlg.FileName, true);
                    }
                    else if(HPGLExt.IndexOf(ext) != -1)
                    {
                        FCADImage = new HPGLImage();
                        LoadFile(openFileDlg.FileName, true);
                    }
                    else
                    {
                        FCADImage = new CADRasterImage();
                        LoadFile(openFileDlg.FileName, false);
                    }
                }
            }
            //CADImage.isImageDraw = false;
            if(this.BackColor == Color.White)
                FCADImage.DefaultColor = Color.Black;
            else FCADImage.DefaultColor = Color.White;
            FCADImage.UseWinEllipse = false;
            Orb3D.CADImage = FCADImage;
            Orb3D.Visible = false;
            this.tlbTool.Buttons[3].Pushed = false;
            Orb3D.Disable3dOrbit();
            deactiv = false;
            Cursor.Current = Cursors.Default;
            ViewLayouts();
            this.cbLayouts.Enabled = true;
            DoResize(true);
            curFileName = openFileDlg.FileName;
            AddNewCADImage();
            //clear readline objects
            rlObjects.Clear();
            rlPoints.Clear();
        }

        private void LoadFile(string fileName, bool det3dBt)
        {
            this.tlbTool.Buttons[2].Enabled = det3dBt;
            this.tlbTool.Buttons[3].Enabled = det3dBt;
            FCADImage.LoadFromFile(fileName);
        }

        private void AddNewCADImage()
        {
            ScaleData d1 = new ScaleData();
            if (!cadImageList.ContainsKey(curFileName))
            {
                MenuItem m1 = new MenuItem();
                m1.Text = curFileName;
                m1.Click += new System.EventHandler(MenuClick);
                this.mainMenu.MenuItems[0].MenuItems.Add(m1);
                cadImageList.Add(curFileName, FCADImage);
                d1.position = pos;
                d1.scale = scale;
                scaleCoeff.Add(curFileName, d1);
            }
            else
            {
                d1.position = pos;
                d1.scale = scale;
                scaleCoeff[curFileName] = d1;
            }
        }

        private void MenuClick(object sender, EventArgs e)
        {
            SelectItem((sender as MenuItem).Text);
            if (this.BackColor == Color.White)
                FCADImage.DefaultColor = Color.Black;
            else FCADImage.DefaultColor = Color.White;
            this.Invalidate();
        }

        private void SaveCADImageScale()
        {
            if (cadImageList.ContainsKey(curFileName))
            {
                ScaleData d1 = new ScaleData();
                d1.position = pos;
                d1.scale = scale;
                scaleCoeff[curFileName] = d1;
            }
        }

        private void ApplyCADImageScale()
        {
            if(cadImageList.ContainsKey(curFileName))
            {
                ScaleData d1 = (ScaleData)scaleCoeff[curFileName];
                pos = d1.position;
                old_Pos = pos;
                scale = d1.scale;
            }
        }
        
        public void DoResize(bool detinv)
        {
            if((FCADImage == null)||(deactiv)) return;
            wh = (float)(FCADImage.AbsWidth / FCADImage.AbsHeight);
            float new_wh = (float)this.ClientRectangle.Width / (float)this.ClientRectangle.Height;
            curClRect = this.ClientRectangle;
            if (new_wh > wh)
                curClRect.Width = (int)(curClRect.Height * wh);
            else
            {
                if (new_wh < wh)
                    curClRect.Height = (int)(curClRect.Width / wh);
                else
                    curClRect = this.ClientRectangle;
            }
            if (detinv)
            {
                pos.X = (this.ClientRectangle.Width - curClRect.Width) / 2;
                pos.Y = (this.ClientRectangle.Height - curClRect.Height) / 2;
                this.Invalidate();
            }
        }

        public void DoCoeff()
        {
            if(FCADImage == null) return;
            wh = (float)(FCADImage.AbsWidth / FCADImage.AbsHeight);
            float new_wh = (float)this.ClientRectangle.Width / (float)this.ClientRectangle.Height;
            curClRect = this.ClientRectangle;
            if (new_wh > wh)
                curClRect.Width = (int)(curClRect.Height * wh);
            else
            {
                if (new_wh < wh)
                    curClRect.Height = (int)(curClRect.Width / wh);
                else
                    curClRect = this.ClientRectangle;
            }
        }

        public void White_Click()
        {
            this.BackColor = Color.White;
            FCADImage.DefaultColor = Color.Black;
        }

        public void Black_Click()
        {
            this.BackColor = Color.Black;
            FCADImage.DefaultColor = Color.White;
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            Protection.Register();
            wh = 1f;
            curClRect = this.ClientRectangle;
            if (this.argFileName != null)
            {
                openFileDlg.FileName = this.argFileName;
                LoadFile(false);
            }
        }

        private void MainForm_Resize(object sender, EventArgs e)
        {
            DoResize(true);
        }

        private void MainForm_Paint(object sender, PaintEventArgs e)
        {
           SetEnableControls(false);
           DrawCADImage(e.Graphics);
           SetEnableControls(true);
           //draw readline new object
           if (this.onReadLine)
           {
               if (rlPoints.Count > 1)
                   e.Graphics.DrawLines(new Pen(Color.Red), (Point[])rlPoints.ToArray(typeof(Point)));
           }
        }

        private void SetEnableControls(bool val)
        {
            this.mainMenu.MenuItems[0].Enabled = val;
            this.mainMenu.MenuItems[1].Enabled = val;
            this.mainMenu.MenuItems[0].MenuItems[1].Enabled = val;
            this.tlbTool.Buttons[3].Enabled = val;
        }

        public void DrawCADImage(Graphics gr)
        {
            if (FCADImage == null) return;
            try
            {
                deactiv = false;
                Shift();
                Rectangle tmp = new Rectangle(pos.X, pos.Y, (int)(curClRect.Width * scale), (int)(curClRect.Height * scale));
                FCADImage.Draw(gr, tmp, this.ClientRectangle);
            }
            catch
            {
                return;
            }
        }

        public void Shift()
        {
            pos.X = (int)(old_Pos.X - (old_Pos.X - pos.X) * scale / prev_scale);
            pos.Y = (int)(old_Pos.Y - (old_Pos.Y - pos.Y) * scale / prev_scale);
            prev_scale = scale;
        }

        private void MainForm_MouseUp(object sender, MouseEventArgs e)
        {
            det1 = false;
        }

        private void MainForm_MouseMove(object sender, MouseEventArgs e)
        {
            if (FCADImage == null) return;
            //add readline poly point and draw current object readline
            if (this.onReadLine)
            {
                rlPoints.Add(new Point(e.X, e.Y));
                if (rlPoints.Count > 1)
                {
                    Graphics gr = this.CreateGraphics();
                    gr.DrawLines(new Pen(Color.Red), (Point[])rlPoints.ToArray(typeof(Point)));
                }
            }
            else
            {
                //move image
                if ((det1) && (!Orb3D.Visible))
                {
                    pos.X -= (cX - e.X);
                    pos.Y -= (cY - e.Y);
                    cX = e.X;
                    cY = e.Y;
                    this.Invalidate();
                }
            }
            old_Pos = new Point(e.X, e.Y);
        }

        private void MainForm_MouseDown(object sender, MouseEventArgs e)
        {
            cX = e.X;
            cY = e.Y;
            det1 = true;
            #if Import
            AddNewRdlnObject();
            #endif
        }

        //readline add new polyline
        #if Import
        private void AddNewRdlnObject()
        {
            if (rlPoints.Count > 0)
            {
                CADPolyLine poly = new CADPolyLine();
                CADVertex vert0;
                for (int i = 0; i < rlPoints.Count; i++)
                {
                    vert0 = new CADVertex();
                    vert0.Point = GetRealPointUsingsgImagePoint(((Point)rlPoints[i]).X, ((Point)rlPoints[i]).Y);
                    poly.AddEntity(vert0);
                }
                poly.Color = Color.Red;
                poly.LWeight = 1;
                FCADImage.Converter.GetSection(ConvSection.Entities).AddEntity(poly);
                rlObjects.Add(poly.GetHashCode());
                FCADImage.Converter.OnCreate(poly);
                FCADImage.Converter.Loads(poly);
                FCADImage.GetExtents();
                DoResize(false);
                rlPoints.Clear();
            } 
        }
        #endif

        private DPoint GetRealPointUsingsgImagePoint(float X, float Y)
        {
            Rectangle tmp = new Rectangle();
            tmp = curClRect;
            DRect Rect = new DRect(tmp.Left, tmp.Top,
                tmp.Right, tmp.Bottom);
            float fKoef = (float)(FCADImage.AbsHeight / FCADImage.AbsWidth);
            Rect.left = Rect.left * scale + pos.X;
            Rect.right = Rect.right * scale + pos.X;
            Rect.top = Rect.top * scale + pos.Y;
            Rect.bottom = Rect.top + (Rect.right - Rect.left) * fKoef;
            ScaleRect.X = FCADImage.AbsWidth / (Rect.right - Rect.left);
            ScaleRect.Y = FCADImage.AbsHeight / (Rect.bottom - Rect.top);
            DPoint mousePt = new DPoint(0, 0, 0);
            DPoint newmousePt = new DPoint(0, 0, 0);
            DPoint pX, pY, pCenter;
            DRect DRectExtentsCAD = FCADImage.Extents;
            mousePt.X = DRectExtentsCAD.left + (X - pos.X) * ScaleRect.X;
            mousePt.Y = DRectExtentsCAD.top - (Y - pos.Y) * ScaleRect.Y;
            pCenter.X = 0.5f * (DRectExtentsCAD.right + DRectExtentsCAD.left);
            pCenter.Y = 0.5f * (DRectExtentsCAD.top + DRectExtentsCAD.bottom);
            pCenter.Z = 0.5f * (DRectExtentsCAD.z2 + DRectExtentsCAD.z1);
            pX.X = 1; pX.Y = 0; pX.Z = 0;
            pY.X = 0; pY.Y = 1; pY.Z = 0;
            pX = FCADImage.GetRealImagePoint(pX);
            pY = FCADImage.GetRealImagePoint(pY);
            MoveToPosition(ref mousePt, pCenter, -1);
            newmousePt.X = mousePt.X * pX.X + mousePt.Y * pX.Y;
            newmousePt.Y = mousePt.X * pY.X + mousePt.Y * pY.Y;
            MoveToPosition(ref newmousePt, pCenter, +1);
            return newmousePt;
        }

        private void MoveToPosition(ref DPoint point, DPoint aPos, int direction)
        {
            point.X = point.X + direction * aPos.X;
            point.Y = point.Y + direction * aPos.Y;
            point.Z = point.Z + direction * aPos.Z;
        }

        public void Go3dOrbit()
        {
            Orb3D.Visible = !Orb3D.Visible;
            this.tlbTool.Buttons[3].Pushed = ! this.tlbTool.Buttons[3].Pushed;
            det1 = false;
            if (Orb3D.Visible) Orb3D.Enable3dOrbit();
            else Orb3D.Disable3dOrbit();
            this.Invalidate();
        }

        public void DoZoomIn()
        {
            if (FCADImage == null) return;
            old_Pos = new Point(this.ClientRectangle.Width / 2, this.ClientRectangle.Height / 2);
            scale = scale * 2;
            this.Invalidate();
        }

        public void DoZoomOut()
        {
            if (FCADImage == null) return;
            old_Pos = new Point(this.ClientRectangle.Width / 2, this.ClientRectangle.Height / 2);
            scale = scale / 2;
            this.Invalidate();
        }

        private void miAbout_Click(object sender, EventArgs e)
        {
            this.abForm1.ShowDialog();
            this.Invalidate();
        }

        private void miReg_Click(object sender, EventArgs e)
        {
            regForm1.ShowDialog();
            this.Invalidate();
        }

        private void miClose_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void miSave_Click(object sender, EventArgs e)
        {
            SaveAsImage();
        }

        public void SaveAsImage()
        {
            if (FCADImage == null) return;
            if (saveImgDlg.ShowDialog() != DialogResult.OK) return;
            this.Invalidate();
            DRect tmpRect = new DRect(0, 0, curClRect.Width * scale, curClRect.Height * scale);
            if (saveImgDlg.FileName != null)
            {
                string tmp = saveImgDlg.FileName;
                if (tmp.ToUpper().IndexOf(".BMP") != -1)
                    FCADImage.SaveToFile(saveImgDlg.FileName, ImageFormat.Bmp, tmpRect);
            }
            if (!(FCADImage is CADRasterImage))
                this.DoResize(true);
        }

        private void MainForm_Deactivate(object sender, EventArgs e)
        {
            deactiv = true;
        }

        public void ResetScaling()
        {
            scale = 1.0f;
            prev_scale = 1;
            pos.X = (this.ClientRectangle.Width - curClRect.Width) / 2;
            pos.Y = (this.ClientRectangle.Height - curClRect.Height) / 2;
        }

        private void SelectItem(string name)
        {
           SaveCADImageScale();
           FCADImage = (CADImage)cadImageList[name];
           if (FCADImage is RasterImage.CADRasterImage)
           {
               this.tlbTool.Buttons[2].Enabled = false;
               this.tlbTool.Buttons[3].Enabled = false;
           }
           else
           {
               this.tlbTool.Buttons[2].Enabled = true;
               this.tlbTool.Buttons[3].Enabled = true;
           }
           Orb3D.CADImage = FCADImage;
           ViewLayouts();
           curFileName = name;
           ApplyCADImageScale();
           DoCoeff();
        }

        private void RemoveItem(string name)
        {
            if (cadImageList.Count == 0) return;
            int i;
            for (i = 2; i < this.mainMenu.MenuItems[0].MenuItems.Count; i++)
            {
                if (this.mainMenu.MenuItems[0].MenuItems[i].Text == name)
                    break;
            }
            int ind = i - 2;
            this.mainMenu.MenuItems[0].MenuItems.RemoveAt(ind + 2);
            cadImageList.Remove(name);
            scaleCoeff.Remove(name);
            if (cadImageList.Count > ind)
            {
                curFileName = this.mainMenu.MenuItems[0].MenuItems[ind + 2].Text;
                SelectItem(curFileName);
            }
            else
            {
                if (ind - 1 > 0)
                {
                    curFileName = this.mainMenu.MenuItems[0].MenuItems[ind + 1].Text;
                    SelectItem(curFileName);
                }
                else
                {
                    curFileName = "";
                    if(FCADImage != null)
                        FCADImage.Dispose();
                    FCADImage = null;
                    this.Invalidate();
                }
            }
            ViewLayouts();
        }

        private void miCloseFile_Click(object sender, EventArgs e)
        {
            RemoveItem(curFileName);
        }

        private void tlbTool_ButtonClick(object sender, ToolBarButtonClickEventArgs e)
        {
            #if Import
            AddNewRdlnObject();
            #endif
            switch(this.tlbTool.Buttons.IndexOf(e.Button))
            {
                case 0:
                    DoZoomIn();
                    break;
                case 1:
                    DoZoomOut();
                    break;
                case 2:
                    SaveAsImage();
                    break;
                case 3:
                    Go3dOrbit();
                    break;
                case 4:
                    ChangeBackGroung(this.BackColor);
                    break;
                case 5:
                    SelectReadline();
                    break;
                case 6:
                    #if Import
                    RemoveAllReadlineObjects();
                    #endif
                    break;
            }
        }

        private void SelectReadline()
        {
            onReadLine = !onReadLine;
            this.tlbTool.Buttons[5].Pushed = onReadLine;
        }

        #if Import
        private void RemoveAllReadlineObjects()
        {
            int j = FCADImage.Converter.Entities.Count - this.rlObjects.Count;
            this.rlObjects.Clear();
            for (int i = FCADImage.Converter.Entities.Count - 1; i >= j; i--)
                FCADImage.Converter.Entities.Remove(i);
            this.Invalidate();
        }
        #endif

        private void ChangeBackGroung(Color backColor)
        {
            this.tlbTool.Buttons[4].Pushed = !this.tlbTool.Buttons[4].Pushed;
            if (this.BackColor == Color.Black)
            {
                if(FCADImage != null)
                   FCADImage.DefaultColor = Color.Black;
                this.BackColor = Color.White;
            }
            else
            {
                if(FCADImage != null)
                    FCADImage.DefaultColor = Color.White;
                this.BackColor = Color.Black;
            }
        }

        public void ViewLayouts()
        {
            int I;
            if (FCADImage == null)
            {
                this.cbLayouts.Items.Clear();
                this.cbLayouts.Enabled = false;
                return;
            }
            cbLayouts.Items.Clear();
            for (I = 0; I < FCADImage.LayoutsCount; I++)
                cbLayouts.Items.Add(FCADImage.GetLayoutName(I));
            cbLayouts.SelectedIndex = FCADImage.DefaultLayoutIndex;
            FCADImage.SetCurrentLayout(cbLayouts.SelectedIndex);
        }

        private void cbLayouts_SelectedIndexChanged(object sender, System.EventArgs e)
        {
            ChangeCurrLayoutIndex();
        }

        public void ChangeCurrLayoutIndex()
        {
            FCADImage.SetCurrentLayout(cbLayouts.SelectedIndex);
            DoResize(true);
            ResetScaling();
            if (cbLayouts.SelectedIndex != 0)
            {
                Orb3D.Visible = false;
                Orb3D.Disable3dOrbit();
            }
        }

        private void miDefault_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotDefault();
            this.Invalidate();
        }

        private void miTop_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.Top);
            this.Invalidate();
        }

        private void miBottom_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.Bottom);
            this.Invalidate();
        }

        private void miLeft_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.Left);
            this.Invalidate();
        }

        private void miRight_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.Right);
            this.Invalidate();
        }

        private void miFront_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.Front);
            this.Invalidate();
        }

        private void miBack_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.Back);
            this.Invalidate();
        }

        private void miSwIso_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.SWIsometric);
            this.Invalidate();
        }

        private void miSeIso_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.SEIsometric);
            this.Invalidate();
        }

        private void miNeIso_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.NEIsometric);
            this.Invalidate();
        }

        private void miNwIso_Click(object sender, EventArgs e)
        {
            if (FCADImage == null) return;
            FCADImage.RotToView(CADViewDirection.NWIsometric);
            this.Invalidate();
        }
    }
}
